#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ranlib.h"
#include "util.h"

/* Cumulative Normal distribution from Hull */
double cumNorm(double x){
  const double a1 = 0.319381530,
    a2 =-0.356563782,
    a3 = 1.781477937,
    a4 =-1.821255978,
    a5 = 1.330274429,
    gmma = 0.2316419;
  double k,nprime;

  if(x<0){
    return(1 - cumNorm(-x));
  }else{
    k = 1.0/(1.0 + gmma*x);
    nprime = exp(-x*x/2.0)/sqrt(2*3.141592654);
    return(1 - nprime*( a1*k
                        +a2*k*k
                        +a3*k*k*k
                        +a4*k*k*k*k
                        +a5*k*k*k*k*k));
  }
}



double sgn(double x){
  if(x >= 0.0){
    return(1.0);
  }else{
    return(-1.0);
  }
}


/* Bivariate normal distribution 

   Inputs are upper limits a and b correlation rho between them.
   Formula from Hull p 260 to four decimal places 
   and 
   Z. Drezner, "Computation of Bivariate Normal Integral",
   Mathematics of Computation, vol 32, no 141, Jan 1978, pp 277-279.

   Checked with three different case from Kolb.
   Note that the following is slightly different from what one may
   infer from Hull. Where Hull may have suggested > instead of >= in
   someplaces we seem to agree with Kolb only i we have the = in place.
   See code below and also look at testbi.c for test cases.
*/
double BivarCND(double a, double b, double rho){
  int i, j;
  double A[4], B[4], sum, ap, bp, f;
  double rho1, rho2, delta;
  double cumNorm(double x);
 
  if( a * b * rho > 0){
    rho1 = sgn(a) * ( rho * a - b) / sqrt(a*a - 2.0*rho*a*b + b*b);
    rho2 = sgn(b) * ( rho * b - a) / sqrt(a*a - 2.0*rho*a*b + b*b);
    delta = (1.0 - sgn(a) * sgn(b))/4.0;
    sum = BivarCND(a,0.0,rho1) + BivarCND(b,0.0,rho2) - delta;
  }else{
    if(b <= 0 && a >= 0 && rho >= 0) {
      sum = cumNorm(b)-BivarCND(-a,b,-rho);
    }else if(a <=0 && b >=0 && rho >= 0){
      sum = cumNorm(a)-BivarCND(a,-b,-rho);
    }else if(rho <= 0.0 && a >= 0 && b >= 0){
      sum = cumNorm(a) + cumNorm(b) - 1.0 + BivarCND(-a,-b, rho); 
    }else{
      /*
        if( a <= 0.0 && b <= 0.0 && rho <=0.0 ){
      */
      /* The following if a <=0, b<=0, rho<=0 */
      A[0] = 0.3253030;
      A[1] = 0.4211071;
      A[2] = 0.1334425;
      A[3] = 0.006374323;
                  
      B[0] = 0.1337764;
      B[1] = 0.6243247;
      B[2] = 1.3425378;
      B[3] = 2.2626645;

      ap = a / sqrt(2.0 * (1.0 - rho*rho));
      bp = b / sqrt(2.0 * (1.0 - rho*rho));

      sum = 0.0;
      for(i=0;i<4;i++){
        for(j=0;j<4;j++){
          f = exp(ap * (2.0 * B[i] - ap) + bp*(2.0 * B[j] - bp) + \
                  2.0 * rho * (B[i] - ap) * (B[j] - bp));
          sum += A[i] * A[j] * f;
        }
      }
      sum *= sqrt(1.0 - rho*rho);
      sum /= 3.141592654;
    }
  }
  return(sum);
} /* End BivarCND */


/* Black-Scholes european call pricing with continuous dividends from Tucker
 *
 * S = current price
 * X = strike price
 * r = risk-free rate
 * s = sigma = standard deviation
 * T = periods to maturity
 * d = delta = dividend rate
 */
#define N(x) cumNorm((x));

double euroCall(double S, double X, double r, double s, double T, double d){
  double d3,Sd;
  double n1,n2;

  Sd = S*exp(-d*T);
  d3 = (log(Sd/X) + (r + (s*s/2))*T)/(s*sqrt(T));
  n1 = N(d3);
  n2 = N(d3 - s*sqrt(T));

  return( Sd*n1 - X*exp(-r*T)*n2 );
}


/************************************************************************
 control3.c
 M. Huq
 Set of routines written to evaluate the max of two risky assets based on
 Stulz 1982 (Journal of Financial Economics, vol 10, 161-185 (1982)).
 *************************************************************************/
 /* 
   Cumulative bivariate normal distribution
   N_2(alpha, beta, theta)
   upper limits of integration alpha and beta and correlation theta.

   WHAT DO WE DO HERE? Need case where theta is non-zero. Hence cannot
   do trivial splitting. Here then, we have to diagonalize the covariance
   matrix and then use the product of the Cumulative normal dist in each
   direction. Have to be careful here when we form the matrix based on the
   expressions given in the paper.
 */

/************************************************************************
 In the following we use the following notation (from paper)

 M := European call option price on the min of two underlying stocks.
 MX:= European call option price on the max of two underlying stocks.
 V := Stock price of underlier one
 H := Stock price of underlier two
 rhovh:= correlation between V and H
 sigmav:= volatility of V
 sigmah:= volatility of H
 R := riskfree interest rate
 tau := T-t
 F := Exercise price
 delta := dividend rate. But note this is not taken into account for
          valuation. What is the correct modification here? <--question for
          Scott and Scott!

 *************************************************************************/
double stulzMax(double V, double H, double sigmav, double sigmah, \
          double rhovh, double R, double F, double tau,double delta)
{
  double M, MX, alpha1, alpha2, beta1, beta2, gamma1, gamma2, sigma, rhoc;

  /* Define secondary variables */
  sigma = sqrt(sigmav*sigmav + sigmah*sigmah - 2.0 * rhovh*sigmav*sigmah);

  rhoc = (rhovh * sigmav - sigmah) / sigma;

  gamma1 = (log(H/F) + (R-0.5*sigmah*sigmah)*tau)/(sigma * sqrt(tau));
  gamma2 = (log(V/F) + (R-0.5*sigmav*sigmav)*tau)/(sigma * sqrt(tau));

  alpha1 = gamma1 + sigmah * sqrt(tau);
  beta1  = gamma2 + sigmav * sqrt(tau);

  alpha2 = (log(V/H) - 0.5 * sigma*sigma*sqrt(tau))/(sigma * sqrt(tau));
  beta2  = (log(H/V) - 0.5 * sigma*sigma*sqrt(tau))/(sigma * sqrt(tau));

  /* Evaluate the European call on the min of two risky stocks using
     equation 11. See form of the same page 166 (first eqn in section 3).
  */ 
  M = H * BivarCND(alpha1, alpha2, rhoc) \
    + V * BivarCND(beta1, beta2, rhoc)   \
    - F * exp(-R * tau) * BivarCND(gamma1,gamma2, rhovh);

  /* Now with M compute the European call option on the max of two
     underliers using now equation 12 from Stulz. Note that this formula
     required the pricing of European call options on single assets. For
     this I used the Black-Scholes formula already in the code. */
  MX = euroCall( V,  F,  R,  sigmav,  tau,  delta) \
     + euroCall( H,  F,  R,  sigmav,  tau,  delta) \
     - M;

  return(MX);
} /* End stulzMax */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* The following is from  a formula in Kolb, "Futures, Options and Swaps"
   An independent description of what is in Stulz above. 
   Gives the same answer as Kolb page 603. Verified.
*/
double maxOf2Risk(double S1, double S2, double X, double tau, \
       double sigma1, double sigma2, double r, double delta1, \
       double delta2, double rho){
  double Sigma, rho1, rho2, w1, w2, w3, w4, Q1, Q2, Q3;

  Sigma = sqrt(sigma1*sigma1 + sigma2*sigma2 - 2.0 * rho*sigma1*sigma2);
  rho1 = (rho * sigma2 - sigma1)/Sigma;
  rho2 = (rho * sigma1 - sigma2)/Sigma;

  w1 = log(S1/X) + (r - delta1 + 0.5 * sigma1*sigma1)*tau;
  w1 /= sigma1 * sqrt(tau);

  w2 = log(S2/X) + (r - delta2 + 0.5 * sigma2*sigma2)*tau;
  w2 /= sigma2 * sqrt(tau);

  w3 = log(S1/S2) + (delta2 - delta1 + 0.5 * Sigma*Sigma)*tau;
  w3 /= Sigma * sqrt(tau);

  w4 = log(S2/S1) + (delta1 - delta2 + 0.5 * Sigma*Sigma)*tau;
  w4 /= Sigma * sqrt(tau);

  Q1 = S1 * exp(-delta1*tau) * \
      (cumNorm(w3) - BivarCND(-w1,w3,rho1));
  Q2 = S2 * exp(-delta2*tau) * \
      (cumNorm(w4) - BivarCND(-w2,w4,rho2));
  Q3 = X * exp(-r*tau) * \
      BivarCND(-w1 + sigma1*sqrt(tau), -w2+sigma2*sqrt(tau), rho);

  return(Q1 + Q2 + Q3 - X * exp(-r*tau));    
}


double ceuropean(int underliers, double *price, double strike, double interest, double variance, double dividend,
		 double tolerance){

    double *parm, *scratch, *dam, *covariance, *growth, error, *newprice,f,g,beta,G,
	answer,count=0.0,w;
    double S1=0.0,S2=0.0,S3=0.0,S4=0.0,S5=0.0,S6=0.0,S7=0.0,S8=0.0,S9=0.0,S10=0.0,S11=0.0,S12=0.0,S13=0.0,S14=0.0,
	S15=0.0,S16=0.0,S17=0.0,num,den;
    int i,j,k,max1, max2;

  
    allocate1dDouble(&parm, underliers*(underliers+3)/2+1);
    allocate1dDouble(&scratch, underliers);
    allocate1dDouble(&dam, underliers);
    allocate1dDouble(&covariance, underliers*underliers);
    allocate1dDouble(&growth, underliers);
    allocate1dDouble(&newprice, underliers);

    if(price[0] > price[1]){
	max1 = 0;
	max2 = 1;
    }else{
	max1 = 1;
	max2 = 0;
    }

    for(i=2; i<underliers; ++i){
	if(price[i] > price[max1]){
	    max2 = max1;
	    max1 = i;
	}else if(price[i] > price[max2]){
	    max2 = i;
	}
    }
  
    G = maxOf2Risk(price[max1],
		   price[max2],
		   strike,
		   1.0,
		   sqrt(variance),
		   sqrt(variance),
		   interest,
		   dividend,
		   dividend,
		   0.0);

    k = 0;
    for(i=0; i<underliers; ++i){
	for(j=0; j<underliers; ++j){
	    if(i==j){
		covariance[k] = variance;
		dam[i] = interest - dividend - variance/2.0;
	    }else{
		covariance[k] = 0;
	    }
	    k++;
	}
    }

    setgmn(dam, covariance, underliers, parm);

    do{
	for(i=0; i<50; ++i){
	    genmn(parm, growth, scratch);

	    newprice[0] = price[0]*exp(growth[0]);
	    f = newprice[0];
	    for(j=1; j<underliers; j++){
		newprice[j] = price[j]*exp(growth[j]);
		if(newprice[j] > f) f = newprice[j];
	    }
	    f = dMax(0, exp(-interest)*(f - strike));
      
	    g = dMax(newprice[max1],newprice[max2]);
	    g = dMax(0, exp(-interest)*(g - strike));

	    count += 1.0;

	    w = 1.0;
	    S1 += w; 
	    S2 += w*w; 
	    S7 += w*f; 
	    S8 += w*(g-G); 
	    S9 += w*w*f; 
	    S10 += w*w*(g-G); 
	    S11 += w*w*f*f; 
	    S12 += w*w*f*(g-G); 
	    S13 += w*w*(g-G)*(g-G); 
	    S14 += w*f*(g-G); 
	    S15 += w*(g-G)*(g-G); 
	    S16 += w*f*f; 
	    S17 += w * g; 
	}
	num = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1));
	den =  (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));
	beta =num/den; 
	/* 			beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/ */
	/*       (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)); */
    
	answer = (S7-beta*S8)/S1;
    
	error = sqrt(1/(S1*S1)*((S11 -2*beta*S12 + beta*beta*S13)   
				-(2/S1)*(S7-beta*S8)*(S9-beta*S10)  
				+(S2/(S1*S1))*(S7*S7 -2*beta*S7*S8 +beta*beta*S8*S8) )); 

    }while(fabs(error/answer) > tolerance);

    deallocate1dDouble(&parm, underliers*(underliers+3)/2+1);
    deallocate1dDouble(&scratch, underliers);
    deallocate1dDouble(&dam, underliers);
    deallocate1dDouble(&covariance, underliers*underliers);
    deallocate1dDouble(&growth, underliers);
    deallocate1dDouble(&newprice, underliers);
    /*
	 printf("count = %f\tanswer =  %f \tden = %f \tnum= %f \n",count,answer,den,num); 
*/
    return(answer);
}


double european(int underliers, double *price, double strike, double interest, double variance, double dividend,
		double tolerance){

    double *parm, *scratch, *dam, *covariance, *growth, S1=0.0,S2=0.0,S3=0.0,S4=0.0,S5=0.0, error, *newprice,f,g,beta,G,
	answer,count=0.0;
    int i,j,k,max1, max2;

    allocate1dDouble(&parm, underliers*(underliers+3)/2+1);
    allocate1dDouble(&scratch, underliers);
    allocate1dDouble(&dam, underliers);
    allocate1dDouble(&covariance, underliers*underliers);
    allocate1dDouble(&growth, underliers);
    allocate1dDouble(&newprice, underliers);

    k = 0;
    for(i=0; i<underliers; ++i){
	for(j=0; j<underliers; ++j){
	    if(i==j){
		covariance[k] = variance;
		dam[i] = interest - dividend - variance/2.0;
	    }else{
		covariance[k] = 0;
	    }
	    k++;
	}
    }

    setgmn(dam, covariance, underliers, parm);

    do{
	for(i=0; i<50; ++i){
	    genmn(parm, growth, scratch);

	    f = newprice[0] = price[0]*exp(growth[0]);
	    for(j=1; j<underliers; j++){
		newprice[j] = price[j]*exp(growth[j]);
		if(newprice[j] > f) f = newprice[j];
	    }
	    f = dMax(0, exp(-interest)*(f - strike));
      
	    count += 1.0;

	    S1 += f;
	    S2 += f*f;
	}

	answer = S1/count;
	error = sqrt((S2/count - S1/count*S1/count)/count);
    
    }while(fabs(error/answer) > tolerance);

    deallocate1dDouble(&parm, underliers*(underliers+3)/2+1);
    deallocate1dDouble(&scratch, underliers);
    deallocate1dDouble(&dam, underliers);
    deallocate1dDouble(&covariance, underliers*underliers);
    deallocate1dDouble(&growth, underliers);
    deallocate1dDouble(&newprice, underliers);

    return(answer);
}
